# coding=utf-8
"""Model DOE-2 Properties."""
from ladybug_geometry.geometry3d import Face3D
from honeybee.units import parse_distance_string

from honeybee_doe2.grouping import _grouped_floor_boundary


class ModelDoe2Properties(object):
    """DOE-2 Properties for Dragonfly Model.

    Args:
        host: A dragonfly_core Model object that hosts these properties.

    Properties:
        * host
    """
    # dictionary mapping validation error codes to a corresponding check function
    ERROR_MAP = {
        '030101': 'check_room_2d_floor_plate_vertex_count',
        '030102': 'check_no_room_2d_floor_plate_holes',
        '030103': 'check_no_story_courtyards'
    }

    def __init__(self, host):
        """Initialize ModelDoe2Properties."""
        self._host = host

    @property
    def host(self):
        """Get the Model object hosting these properties."""
        return self._host

    def check_for_extension(self, raise_exception=True, detailed=False):
        """Check that the Model is valid for DOE-2 simulation.

        This process includes all relevant dragonfly-core checks as well as checks
        that apply only for DOE-2.

        Args:
            raise_exception: Boolean to note whether a ValueError should be raised
                if any errors are found. If False, this method will simply
                return a text string with all errors that were found. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A text string with all errors that were found or a list if detailed is True.
            This string (or list) will be empty if no errors were found.
        """
        # set up defaults to ensure the method runs correctly
        detailed = False if raise_exception else detailed
        msgs = []
        tol = self.host.tolerance
        ang_tol = self.host.angle_tolerance

        # perform checks for duplicate identifiers, which might mess with other checks
        msgs.append(self.host.check_all_duplicate_identifiers(False, detailed))

        # perform checks for key dragonfly model schema rules
        msgs.append(self.host.check_degenerate_room_2ds(tol, False, detailed))
        msgs.append(self.host.check_self_intersecting_room_2ds(tol, False, detailed))
        msgs.append(self.host.check_plenum_depths(tol, False, detailed))
        msgs.append(self.host.check_window_parameters_valid(tol, False, detailed))
        msgs.append(self.host.check_no_room2d_overlaps(tol, False, detailed))
        msgs.append(self.host.check_collisions_between_stories(tol, False, detailed))
        msgs.append(self.host.check_roofs_above_rooms(tol, False, detailed))
        msgs.append(self.host.check_room2d_floor_heights_valid(False, detailed))
        msgs.append(self.host.check_missing_adjacencies(False, detailed))
        msgs.append(self.host.check_all_room3d(tol, ang_tol, False, detailed))

        # perform checks that are specific to DOE-2
        msgs.append(self.check_room_2d_floor_plate_vertex_count(False, detailed))
        msgs.append(self.check_no_room_2d_floor_plate_holes(False, detailed))
        msgs.append(self.check_no_story_courtyards(tol, False, detailed))

        # output a final report of errors or raise an exception
        full_msgs = [msg for msg in msgs if msg]
        if detailed:
            return [m for msg in full_msgs for m in msg]
        full_msg = '\n'.join(full_msgs)
        if raise_exception and len(full_msgs) != 0:
            raise ValueError(full_msg)
        return full_msg

    def check_generic(self, raise_exception=True, detailed=False):
        """Check generic of the aspects of the Model DOE-2 properties.

        This includes checks for everything except holes in floor plates and
        courtyard stories.

        Args:
            raise_exception: Boolean to note whether a ValueError should be raised
                if any errors are found. If False, this method will simply
                return a text string with all errors that were found.
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A text string with all errors that were found or a list if detailed is True.
            This string (or list) will be empty if no errors were found.
        """
        # set up defaults to ensure the method runs correctly
        detailed = False if raise_exception else detailed
        msgs = []
        # perform checks for specific doe-2 simulation rules
        msgs.append(self.check_room_2d_floor_plate_vertex_count(False, detailed))
        # output a final report of errors or raise an exception
        full_msgs = [msg for msg in msgs if msg]
        if detailed:
            return [m for msg in full_msgs for m in msg]
        full_msg = '\n'.join(full_msgs)
        if raise_exception and len(full_msgs) != 0:
            raise ValueError(full_msg)
        return full_msg

    def check_all(self, raise_exception=True, detailed=False):
        """Check all of the aspects of the Model DOE-2 properties.

        Args:
            raise_exception: Boolean to note whether a ValueError should be raised
                if any errors are found. If False, this method will simply
                return a text string with all errors that were found.
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A text string with all errors that were found or a list if detailed is True.
            This string (or list) will be empty if no errors were found.
        """
        # set up defaults to ensure the method runs correctly
        detailed = False if raise_exception else detailed
        msgs = []
        tol = self.host.tolerance
        # perform checks for specific doe-2 simulation rules
        msgs.append(self.check_room_2d_floor_plate_vertex_count(False, detailed))
        msgs.append(self.check_no_room_2d_floor_plate_holes(False, detailed))
        msgs.append(self.check_no_story_courtyards(tol, False, detailed))
        # output a final report of errors or raise an exception
        full_msgs = [msg for msg in msgs if msg]
        if detailed:
            return [m for msg in full_msgs for m in msg]
        full_msg = '\n'.join(full_msgs)
        if raise_exception and len(full_msgs) != 0:
            raise ValueError(full_msg)
        return full_msg

    def check_room_2d_floor_plate_vertex_count(
            self, raise_exception=True, detailed=False):
        """Check whether any Room2Ds floor geometry exceeds the maximum vertex count.

        The DOE-2 engine currently does not support such rooms and limits the
        total number of vertices to 120.

        Args:
            raise_exception: If True, a ValueError will be raised if the Room2D
                floor plate exceeds the maximum number of vertices supported by
                DOE-2. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        detailed = False if raise_exception else detailed
        msgs = []
        for room in self.host.room_2ds:
            msg = room.properties.doe2.check_floor_plate_vertex_count(False, detailed)
            if detailed:
                msgs.extend(msg)
            elif msg != '':
                msgs.append(msg)
        if detailed:
            return msgs
        full_msg = '\n'.join(msgs)
        if raise_exception and len(msgs) != 0:
            raise ValueError(full_msg)
        return full_msg

    def check_no_room_2d_floor_plate_holes(self, raise_exception=True, detailed=False):
        """Check whether any Room2D floor geometry has holes.

        EQuest currently has no way to represent such rooms so, if the issue
        is not addressed, the hole will simply be removed as part of the
        process of exporting to an INP file.

        Args:
            raise_exception: If True, a ValueError will be raised if the Room2D
                floor plate has one or more holes. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        detailed = False if raise_exception else detailed
        msgs = []
        for room in self.host.room_2ds:
            msg = room.properties.doe2.check_no_floor_plate_holes(False, detailed)
            if detailed:
                msgs.extend(msg)
            elif msg != '':
                msgs.append(msg)
        if detailed:
            return msgs
        full_msg = '\n'.join(msgs)
        if raise_exception and len(msgs) != 0:
            raise ValueError(full_msg)
        return full_msg

    def check_no_story_courtyards(
            self, tolerance=None, raise_exception=True, detailed=False):
        """Check that Story floor plates do not contain courtyards.

        EQuest currently has no way to represent such courtyards so, if the issue
        is not addressed, the courtyards will simply be removed as part of the
        process of exporting to an INP file.

        Args:
            tolerance: The tolerance to be used when joining the Room2D floor
                plates together into a Story floor plate. If None, the Model
                tolerance will be used. (Default: None).
            raise_exception: Boolean to note whether a ValueError should be raised
                if the story contains a courtyard. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        # establish the tolerance and gap width at which point it is clearly a courtyard
        tolerance = self.host.tolerance if tolerance is None else tolerance
        court_width = parse_distance_string('1ft', self.host.units)
        # loop through the stories and identify any courtyards
        story_msgs = []
        for bldg in self.host.buildings:
            for story in bldg.unique_stories:
                floor_geos = [room.floor_geometry for room in story.room_2ds]
                joined_geos = _grouped_floor_boundary(floor_geos, tolerance)
                c_count = 0
                for geo in joined_geos:
                    if geo.has_holes:
                        for hole in geo.holes:
                            try:
                                h_geo = Face3D(hole)
                                h_geo = h_geo.remove_colinear_vertices(court_width)
                                max_len = max(s.length for s in h_geo.boundary_segments)
                                tol_area = max_len * court_width
                                if h_geo.area > tol_area:
                                    c_count += 1
                            except (AssertionError, ValueError):
                                pass  # gap is too small to be a true courtyard
                if c_count != 0:
                    hole_msg = 'a courtyard' if c_count == 1 \
                        else '{} courtyards'.format(c_count)
                    msg = 'The geometry of Story "{}" contains {}, which eQuest ' \
                        'cannot represent.'.format(story.display_name, hole_msg)
                    if detailed:
                        msg = {
                            'type': 'ValidationError',
                            'code': '030103',
                            'error_type': 'Story Floor Plate Contains Courtyards',
                            'extension_type': 'DOE2',
                            'element_type': 'Room2D',
                            'element_id': [r.identifier for r in story.room_2ds],
                            'element_name': [r.display_name for r in story.room_2ds],
                            'message': msg
                        }
                    story_msgs.append(msg)
        if detailed:
            return story_msgs
        if story_msgs != []:
            msg = 'The following Stories have courtyards in their floor plates' \
                ':\n{}'.format('\n'.join(story_msgs))
            if raise_exception:
                raise ValueError(msg)
            return msg
        return ''

    def to_dict(self):
        """Return Model DOE-2 properties as a dictionary."""
        return {'doe2': {'type': 'ModelDoe2Properties'}}

    def apply_properties_from_dict(self, data):
        """Apply the DOE-2 properties of a dictionary to the host Model of this object.

        Args:
            data: A dictionary representation of an entire dragonfly-core Model.
                Note that this dictionary must have ModelDoe2Properties in order
                for this method to successfully apply the DOE-2 properties.
        """
        assert 'doe2' in data['properties'], \
            'Dictionary possesses no ModelDoe2Properties.'
        room_doe2_dicts = []
        if 'buildings' in data and data['buildings'] is not None:
            for b_dict in data['buildings']:
                if 'unique_stories' in b_dict and b_dict['unique_stories'] is not None:
                    for story_dict in b_dict['unique_stories']:
                        for room_dict in story_dict['room_2ds']:
                            try:
                                room_doe2_dicts.append(room_dict['properties']['doe2'])
                            except KeyError:
                                room_doe2_dicts.append(None)
        if len(room_doe2_dicts) != 0:
            for room, r_dict in zip(self.host.room_2ds, room_doe2_dicts):
                if r_dict is not None:
                    room.properties.doe2.apply_properties_from_dict(r_dict)

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return 'Model DOE2 Properties: [host: {}]'.format(self.host.display_name)



# coding=utf-8
"""Room2D DOE-2 Properties."""
from honeybee.typing import float_in_range, float_positive
from honeybee.altnumber import autocalculate
from honeybee_doe2.properties.room import RoomDoe2Properties


class Room2DDoe2Properties(object):
    """DOE-2 Properties for Dragonfly Room2D.

    Args:
        host: A dragonfly_core Room2D object that hosts these properties.
        assigned_flow: A number for the design supply air flow rate for the zone
            the Room2D is assigned to (cfm). This establishes the minimum allowed
            design air flow. Note that the actual design flow may be larger. If
            None, this parameter will not be written into the INP. (Default: None).
        flow_per_area: A number for the design supply air flow rate to
            the zone per unit floor area (cfm/ft2). If None, this parameter
            will not be written into the INP. (Default: None).
        min_flow_ratio: A number between 0 and 1 for the minimum allowable zone
            air supply flow rate, expressed as a fraction of design flow rate.
            Applicable to variable-volume type systems only. If None, this parameter
            will not be written into the INP. (Default: None).
        min_flow_per_area: A number for the minimum air flow per square foot of
            floor area (cfm/ft2). This is an alternative way of specifying the
            min_flow_ratio. If None, this parameter will not be written into
            the INP. (Default: None).
        hmax_flow_ratio: A number between 0 and 1 for the ratio of the maximum
            (or fixed) heating airflow to the cooling airflow. The specific
            meaning varies according to the type of zone terminal. If None, this
            parameter will not be written into the INP. (Default: None).

    Properties:
        * host
        * assigned_flow
        * flow_per_area
        * min_flow_ratio
        * min_flow_per_area
        * hmax_flow_ratio
    """
    __slots__ = (
        '_host', '_assigned_flow', '_flow_per_area', '_min_flow_ratio',
        '_min_flow_per_area', '_hmax_flow_ratio',
    )

    def __init__(
        self, host, assigned_flow=None, flow_per_area=None, min_flow_ratio=None,
        min_flow_per_area=None, hmax_flow_ratio=None
    ):
        """Initialize Room2D DOE-2 properties."""
        # set the main properties of the Room2D
        self._host = host
        self.assigned_flow = assigned_flow
        self.flow_per_area = flow_per_area
        self.min_flow_ratio = min_flow_ratio
        self.min_flow_per_area = min_flow_per_area
        self.hmax_flow_ratio = hmax_flow_ratio

    @property
    def host(self):
        """Get the Room2D object hosting these properties."""
        return self._host

    @property
    def assigned_flow(self):
        """Get or set the design supply air flow rate for the zone (cfm)."""
        return self._assigned_flow

    @assigned_flow.setter
    def assigned_flow(self, value):
        if value is not None:
            value = float_positive(value, 'zone assigned flow')
        self._assigned_flow = value

    @property
    def flow_per_area(self):
        """Get or set the design supply air flow rate per unit floor area (cfm/ft2).
        """
        return self._flow_per_area

    @flow_per_area.setter
    def flow_per_area(self, value):
        if value is not None:
            value = float_positive(value, 'zone flow per area')
        self._flow_per_area = value

    @property
    def min_flow_ratio(self):
        """Get or set the the min supply airflow rate as a fraction of design flow rate.
        """
        return self._min_flow_ratio

    @min_flow_ratio.setter
    def min_flow_ratio(self, value):
        if value is not None:
            value = float_in_range(value, 0.0, 1.0, 'zone min flow ratio')
        self._min_flow_ratio = value

    @property
    def min_flow_per_area(self):
        """Get or set the minimum air flow per square foot of floor area (cfm/ft2)."""
        return self._min_flow_per_area

    @min_flow_per_area.setter
    def min_flow_per_area(self, value):
        if value is not None:
            value = float_positive(value, 'zone min flow per area')
        self._min_flow_per_area = value

    @property
    def hmax_flow_ratio(self):
        """Get or set the ratio of the maximum heating airflow to the cooling airflow.
        """
        return self._hmax_flow_ratio

    @hmax_flow_ratio.setter
    def hmax_flow_ratio(self, value):
        if value is not None:
            value = float_in_range(value, 0.0, 1.0, 'zone heating max flow ratio')
        self._hmax_flow_ratio = value

    def check_floor_plate_vertex_count(self, raise_exception=True, detailed=False):
        """Check whether the Room2D's floor geometry exceeds the maximum vertex count.

        The DOE-2 engine currently does not support such rooms and limits the
        total number of vertices to 120.

        Args:
            raise_exception: If True, a ValueError will be raised if the Room2D
                floor plate exceeds the maximum number of vertices supported by
                DOE-2. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        if len(self.host.floor_geometry.boundary) > 120:
            msg = 'Room2D "{}" has a floor plate with {} vertices, which is more ' \
                'than the maximum 120 vertices supported by DOE-2.'.format(
                    self.host.display_name, len(self.host.floor_geometry.boundary))
            if raise_exception:
                raise ValueError(msg)
            full_msg = self.host._validation_message_child(
                msg, self.host, detailed, '030101', extension='DOE2',
                error_type='Room Exceeds Maximum Vertex Count')
            if detailed:
                return [full_msg]
            if raise_exception:
                raise ValueError(full_msg)
            return full_msg
        return [] if detailed else ''

    def check_no_floor_plate_holes(self, raise_exception=True, detailed=False):
        """Check whether the Room2D's floor geometry has holes.

        EQuest currently has no way to represent such rooms so, if the issue
        is not addressed, the hole will simply be removed as part of the
        process of exporting to an INP file.

        Args:
            raise_exception: If True, a ValueError will be raised if the Room2D
                floor plate has one or more holes. (Default: True).
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).

        Returns:
            A string with the message or a list with a dictionary if detailed is True.
        """
        if self.host.floor_geometry.has_holes:
            hole_count = len(self.host.floor_geometry.holes)
            hole_msg = 'a hole' if hole_count == 1 else '{} holes'.format(hole_count)
            msg = 'Room2D "{}" has a floor plate with {}, which the eQuest ' \
                'interface cannot represent.'.format(self.host.display_name, hole_msg)
            if raise_exception:
                raise ValueError(msg)
            full_msg = self.host._validation_message_child(
                msg, self.host, detailed, '030102', extension='DOE2',
                error_type='Room Contains Holes')
            if detailed:
                return [full_msg]
            if raise_exception:
                raise ValueError(full_msg)
            return full_msg
        return [] if detailed else ''

    @classmethod
    def from_dict(cls, data, host):
        """Create Room2DDoe2Properties from a dictionary.

        Args:
            data: A dictionary representation of Room2DDoe2Properties with the
                format below.
            host: A Room2D object that hosts these properties.

        .. code-block:: python

            {
            "type": 'Room2DDoe2Properties',
            "assigned_flow": 100,  # number in cfm
            "flow_per_area": 1,  # number in cfm/ft2
            "min_flow_ratio": 0.3, # number between 0 and 1
            "min_flow_per_area": 0.3, # number in cfm/ft2
            "hmax_flow_ratio": 0.3  # number between 0 and 1
            }
        """
        assert data['type'] == 'Room2DDoe2Properties', \
            'Expected Room2DDoe2Properties. Got {}.'.format(data['type'])
        new_prop = cls(host)
        auto_dict = autocalculate.to_dict()
        if 'assigned_flow' in data and data['assigned_flow'] != auto_dict:
            new_prop.assigned_flow = data['assigned_flow']
        if 'flow_per_area' in data and data['flow_per_area'] != auto_dict:
            new_prop.flow_per_area = data['flow_per_area']
        if 'min_flow_ratio' in data and data['min_flow_ratio'] != auto_dict:
            new_prop.min_flow_ratio = data['min_flow_ratio']
        if 'min_flow_per_area' in data and data['min_flow_per_area'] != auto_dict:
            new_prop.min_flow_per_area = data['min_flow_per_area']
        if 'hmax_flow_ratio' in data and data['hmax_flow_ratio'] != auto_dict:
            new_prop.hmax_flow_ratio = data['hmax_flow_ratio']
        return new_prop

    def apply_properties_from_dict(self, data):
        """Apply properties from a Room2DDoe2Properties dictionary.

        Args:
            data: A Room2DDoe2Properties dictionary (typically coming from a Model).
        """
        auto_dict = autocalculate.to_dict()
        if 'assigned_flow' in data and data['assigned_flow'] != auto_dict:
            self.assigned_flow = data['assigned_flow']
        if 'flow_per_area' in data and data['flow_per_area'] != auto_dict:
            self.flow_per_area = data['flow_per_area']
        if 'min_flow_ratio' in data and data['min_flow_ratio'] != auto_dict:
            self.min_flow_ratio = data['min_flow_ratio']
        if 'min_flow_per_area' in data and data['min_flow_per_area'] != auto_dict:
            self.min_flow_per_area = data['min_flow_per_area']
        if 'hmax_flow_ratio' in data and data['hmax_flow_ratio'] != auto_dict:
            self.hmax_flow_ratio = data['hmax_flow_ratio']

    def to_dict(self, abridged=False):
        """Return Room2D Doe2 properties as a dictionary."""
        base = {'doe2': {}}
        base['doe2']['type'] = 'Room2DDoe2Properties'
        if self.assigned_flow is not None:
            base['doe2']['assigned_flow'] = self.assigned_flow
        if self.flow_per_area is not None:
            base['doe2']['flow_per_area'] = self.flow_per_area
        if self.min_flow_ratio is not None:
            base['doe2']['min_flow_ratio'] = self.min_flow_ratio
        if self.min_flow_per_area is not None:
            base['doe2']['min_flow_per_area'] = self.min_flow_per_area
        if self.hmax_flow_ratio is not None:
            base['doe2']['hmax_flow_ratio'] = self.hmax_flow_ratio
        return base

    def to_honeybee(self, new_host):
        """Get a honeybee version of this object.

        Args:
            new_host: A honeybee-core Room object that will host these properties.
        """
        return RoomDoe2Properties(
            new_host, self.assigned_flow, self.flow_per_area, self.min_flow_ratio,
            self.min_flow_per_area, self.hmax_flow_ratio
        )

    def from_honeybee(self, hb_properties):
        """Transfer Doe-2 attributes from a Honeybee Room to Dragonfly Room2D.

        Args:
            hb_properties: The RoomDoe2Properties of the honeybee Room that is being
                translated to a Dragonfly Room2D.
        """
        self._assigned_flow = hb_properties._assigned_flow
        self._flow_per_area = hb_properties._flow_per_area
        self._min_flow_ratio = hb_properties._min_flow_ratio
        self._min_flow_per_area = hb_properties._min_flow_per_area
        self._hmax_flow_ratio = hb_properties._hmax_flow_ratio

    def duplicate(self, new_host=None):
        """Get a copy of this object.

        Args:
            new_host: A new Room2D object that hosts these properties.
                If None, the properties will be duplicated with the same host.
        """
        _host = new_host or self._host
        new_room = Room2DDoe2Properties(
            _host, self.assigned_flow, self.flow_per_area, self.min_flow_ratio,
            self.min_flow_per_area, self.hmax_flow_ratio)
        return new_room

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return 'Room2D DOE2 Properties: [host: {}]'.format(self.host.display_name)



# coding: utf-8
"""Write an inp file from a Dragonfly model."""
from __future__ import division

from honeybee_doe2.writer import model_to_inp as hb_model_to_inp


def model_to_inp(
    model, use_multiplier=True, exclude_plenums=False, solve_ceiling_adjacencies=True,
    simulation_par=None, hvac_mapping='Story',
    exclude_interior_walls=False, exclude_interior_ceilings=False, equest_version=None
):
    """Generate an INP string from a Dragonfly Model.

    The resulting string will include all geometry, all fully-detailed constructions
    and materials, all fully-detailed schedules, and the room properties. It will
    also include the simulation parameters. Essentially, the string includes
    everything needed to simulate the model.

    Args:
        model: A dragonfly Model for which an INP representation will be returned.
        use_multiplier: Boolean to note if the multipliers on each Building
            story will be passed along to the generated Honeybee Room objects
            or if full geometry objects should be written for each story in the
            building. (Default: True).
        exclude_plenums: Boolean to indicate whether ceiling/floor plenum depths
            assigned to Room2Ds should generate distinct 3D Rooms in the
            translation. (Default: False).
        solve_ceiling_adjacencies: Boolean to indicate whether adjacencies should
            be solved between interior stories when Room2Ds perfectly match one
            another in their floor plate. This ensures that Surface boundary
            conditions are used instead of Adiabatic ones. (Default: True).
        simulation_par: A honeybee-doe2 SimulationPar object to specify how the
            DOE-2 simulation should be run. If None, default simulation
            parameters will be generated, which will run the simulation for the
            full year. (Default: None).
        hvac_mapping: Text to indicate how HVAC systems should be assigned to the
            exported model. Story will assign one HVAC system for each distinct
            level polygon, Model will use only one HVAC system for the whole model
            and AssignedHVAC will follow how the HVAC systems have been assigned
            to the Rooms.properties.energy.hvac. Choose from the options
            below. (Default: Story).

            * Room
            * Story
            * Model
            * AssignedHVAC

        exclude_interior_walls: Boolean to note whether interior wall Faces
            should be excluded from the resulting string. (Default: False).
        exclude_interior_ceilings: Boolean to note whether interior ceiling
            Faces should be excluded from the resulting string. (Default: False).
        equest_version: An optional text string to denote the version of eQuest
            for which the INP definition will be generated. If unspecified
            or unrecognized, the latest version of eQuest will be used.

    Usage:

    .. code-block:: python

        import os
        from ladybug.futil import write_to_file
        from ladybug_geometry.geometry3d import Point3D, Face3D
        from dragonfly.model import Model
        from dragonfly.building import Building
        from dragonfly.story import Story
        from dragonfly.room2d import Room2D
        from dragonfly.roof import RoofSpecification
        from dragonfly.windowparameter import SimpleWindowRatio
        from honeybee.config import folders

        # Crate an input Model
        pts1 = (Point3D(0, 0, 0), Point3D(10, 0, 0),
                Point3D(10, 10, 0), Point3D(0, 10, 0))
        pts2 = (Point3D(10, 0, 0), Point3D(20, 0, 0),
                Point3D(20, 10, 0), Point3D(10, 10, 0))
        pts3 = (Point3D(0, 0, 3.25), Point3D(20, 0, 3.25),
                Point3D(20, 5, 5), Point3D(0, 5, 5))
        pts4 = (Point3D(0, 5, 5), Point3D(20, 5, 5),
                Point3D(20, 10, 3.25), Point3D(0, 10, 3.25))
        room2d_full = Room2D(
            'R1-full', floor_geometry=Face3D(pts1), floor_to_ceiling_height=4,
            is_ground_contact=True, is_top_exposed=True)
        room2d_plenum = Room2D(
            'R2-plenum', floor_geometry=Face3D(pts2), floor_to_ceiling_height=4,
            is_ground_contact=True, is_top_exposed=True)
        room2d_plenum.ceiling_plenum_depth = 1.0
        roof = RoofSpecification([Face3D(pts3), Face3D(pts4)])
        story = Story('S1', [room2d_full, room2d_plenum])
        story.roof = roof
        story.solve_room_2d_adjacency(0.01)
        story.set_outdoor_window_parameters(SimpleWindowRatio(0.4))
        building = Building('Office_Building_1234', [story])
        model = Model('NewDevelopment1', [building])

        # create the INP string for the model
        inp_str = model.to.inp(model)

        # write the final string into an INP
        inp = os.path.join(folders.default_simulation_folder, 'test_file', 'in.inp')
        write_to_file(inp, inp_str, True)
    """
    # convert the Dragonfly Model to Honeybee
    hb_model = model.to_honeybee(
        'District', use_multiplier=use_multiplier, exclude_plenums=exclude_plenums,
        solve_ceiling_adjacencies=solve_ceiling_adjacencies,
        enforce_adj=False, enforce_solid=True)[0]

    # assign the space polygon geometry to the Honeybee Rooms from Dragonfly
    df_flr_geos = {}  # dictionary to hold the DOE-2 space polygon geometry
    for df_room in model.room_2ds:
        df_flr_geos[df_room.identifier] = df_room.floor_geometry
    for hb_room in hb_model.rooms:
        try:
            hb_room.properties.doe2.space_polygon_geometry = \
                df_flr_geos[hb_room.identifier]
        except KeyError:  # possibly a 3D Room that has no Room2D geometry
            pass

    # patch missing adjacencies to adiabatic in case the models is a sub-selection
    hb_model.properties.energy.missing_adjacencies_to_adiabatic()

    # translate the Honeybee Model to INP
    inp_str = hb_model_to_inp(
        hb_model, simulation_par=simulation_par, hvac_mapping=hvac_mapping,
        exclude_interior_walls=exclude_interior_walls,
        exclude_interior_ceilings=exclude_interior_ceilings,
        equest_version=equest_version)
    return inp_str



"""dragonfly-doe2 commands which will be added to the dragonfly CLI."""
import click

from dragonfly.cli import main
from .translate import translate


# command group for all doe2 extension commands.
@click.group(help='dragonfly doe2 commands.')
def doe2():
    pass


# add sub-commands for doe2
doe2.add_command(translate)

# add doe2 sub-commands to dragonfly CLI
main.add_command(doe2)



"""dragonfly doe2 translation commands."""
import click
import sys
import logging
import json

from ladybug.commandutil import process_content_to_output
from dragonfly.model import Model
from honeybee_doe2.simulation import SimulationPar

from dragonfly_doe2.writer import model_to_inp as writer_model_to_inp

_logger = logging.getLogger(__name__)


@click.group(help='Commands for translating Dragonfly files to DOE-2.')
def translate():
    pass


@translate.command('model-to-inp')
@click.argument('model-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--multiplier/--full-geometry', ' /-fg', help='Flag to note if the '
    'multipliers on each Building story will be passed along to the '
    'generated Honeybee Room objects or if full geometry objects should be '
    'written for each story in the building.', default=True, show_default=True)
@click.option(
    '--plenum/--no-plenum', '-p/-np', help='Flag to indicate whether '
    'ceiling/floor plenum depths assigned to Room2Ds should generate '
    'distinct 3D Rooms in the translation.', default=True, show_default=True)
@click.option(
    '--ceil-adjacency/--no-ceil-adjacency', '-a/-na', help='Flag to indicate '
    'whether adjacencies should be solved between interior stories when '
    'Room2Ds perfectly match one another in their floor plate. This ensures '
    'that Surface boundary conditions are used instead of Adiabatic ones. '
    'Note that this input has no effect when the object-per-model is Story.',
    default=True, show_default=True)
@click.option(
    '--sim-par-json', '-sp', help='Full path to a honeybee-doe2 SimulationPar '
    'JSON that describes all of the settings for the simulation. If unspecified, '
    'default parameters will be generated.', default=None, show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--hvac-mapping', '-hm', help='Text to indicate how HVAC systems should be '
    'assigned to the exported model. Story will assign one HVAC system for each '
    'distinct level polygon, Model will use only one HVAC system for the whole model '
    'and AssignedHVAC will follow how the HVAC systems have been assigned to the'
    'Rooms.properties.energy.hvac. Choose from: Room, Story, Model, AssignedHVAC',
    default='Story', show_default=True, type=str)
@click.option(
    '--include-interior-walls/--exclude-interior-walls', ' /-xw', help='Flag to note '
    'whether interior walls should be excluded from the export.',
    default=True, show_default=True)
@click.option(
    '--include-interior-ceilings/--exclude-interior-ceilings', ' /-xc', help='Flag to '
    'note whether interior ceilings should be excluded from the export.',
    default=True, show_default=True)
@click.option(
    '--equest-version', '-eq', help='Optional text string to denote the version '
    'of eQuest for which the INP definition will be generated. If unspecified '
    'or unrecognized, the latest version of eQuest will be used.',
    default='3.65', show_default=True, type=str)
@click.option(
    '--output-file', '-o', help='Optional INP file path to output the INP string '
    'of the translation. By default this will be printed out to stdout.',
    type=click.File('w'), default='-', show_default=True)
def model_to_inp_cli(
    model_file, multiplier, plenum, ceil_adjacency, sim_par_json, hvac_mapping,
    include_interior_walls, include_interior_ceilings, equest_version, output_file
):
    """Translate a Dragonfly Model file to a Radiance string.

    The resulting string will include all geometry and all modifiers.

    \b
    Args:
        model_file: Full path to a Dragonfly Model JSON or Pkl file.
    """
    try:
        full_geometry = not multiplier
        no_plenum = not plenum
        no_ceil_adjacency = not ceil_adjacency
        exclude_interior_walls = not include_interior_walls
        exclude_interior_ceilings = not include_interior_ceilings
        model_to_inp(
            model_file, full_geometry, no_plenum, no_ceil_adjacency,
            sim_par_json, hvac_mapping,
            exclude_interior_walls, exclude_interior_ceilings,
            equest_version, output_file)
    except Exception as e:
        _logger.exception('Model translation failed.\n{}\n'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def model_to_inp(
    model_file, full_geometry=False, no_plenum=False, no_ceil_adjacency=False,
    sim_par_json=None, hvac_mapping='Story', exclude_interior_walls=False,
    exclude_interior_ceilings=False, equest_version=None, output_file=None,
    multiplier=True, plenum=True, ceil_adjacency=True,
    include_interior_walls=True, include_interior_ceilings=True
):
    """Translate a Model file to an INP string.

    Args:
        model_file: Full path to a Model JSON file (DFJSON) or a Model pkl (DFpkl) file.
        full_geometry: Boolean to note if the multipliers on each Building story
            will be passed along to the generated Honeybee Room objects or if
            full geometry objects should be written for each story in the
            building. (Default: False).
        no_plenum: Boolean to indicate whether ceiling/floor plenum depths
            assigned to Room2Ds should generate distinct 3D Rooms in the
            translation. (Default: False).
        ceil_adjacency: Boolean to indicate whether adjacencies should be solved
            between interior stories when Room2Ds perfectly match one another
            in their floor plate. This ensures that Surface boundary conditions
            are used instead of Adiabatic ones. Note that this input has no
            effect when the object-per-model is Story. (Default: False).
        simulation_par: A honeybee-doe2 SimulationPar object to specify how the
            DOE-2 simulation should be run. If None, default simulation
            parameters will be generated, which will run the simulation for the
            full year. (Default: None).
        hvac_mapping: Text to indicate how HVAC systems should be assigned to the
            exported model. Story will assign one HVAC system for each distinct
            level polygon, Model will use only one HVAC system for the whole model
            and AssignedHVAC will follow how the HVAC systems have been assigned
            to the Rooms.properties.energy.hvac. Choose from the options
            below. (Default: Story).

            * Room
            * Story
            * Model
            * AssignedHVAC

        exclude_interior_walls: Boolean to note whether interior wall Faces
            should be excluded from the resulting string. (Default: False).
        exclude_interior_ceilings: Boolean to note whether interior ceiling
            Faces should be excluded from the resulting string. (Default: False).
        equest_version: An optional text string to denote the version of eQuest
            for which the INP definition will be generated. If unspecified
            or unrecognized, the latest version of eQuest will be used.
        output_file: Optional INP file to output the INP string of the translation.
            If None, the string will be returned from this method. (Default: None).
    """
    # re-serialize the Dragonfly Model
    model = Model.from_file(model_file)

    # load simulation parameters if specified
    sim_par = None
    if sim_par_json is not None:
        with open(sim_par_json) as json_file:
            data = json.load(json_file)
        sim_par = SimulationPar.from_dict(data)

    # create the strings for the model
    multiplier = not full_geometry
    ceil_adjacency = not no_ceil_adjacency
    inp_str = writer_model_to_inp(
        model, multiplier, no_plenum, ceil_adjacency, sim_par, hvac_mapping,
        exclude_interior_walls, exclude_interior_ceilings, equest_version)

    # write out the INP file
    return process_content_to_output(inp_str, output_file)



# coding=utf-8
from dragonfly.properties import ModelProperties, Room2DProperties
import dragonfly.writer.model as model_writer

from .properties.model import ModelDoe2Properties
from .properties.room2d import Room2DDoe2Properties
from .writer import model_to_inp


# set a hidden doe2 attribute on each core geometry Property class to None
# define methods to produce doe2 property instances on each Property instance
ModelProperties._doe2 = None
Room2DProperties._doe2 = None


def model_doe2_properties(self):
    if self._doe2 is None:
        self._doe2 = ModelDoe2Properties(self.host)
    return self._doe2


def room2d_doe2_properties(self):
    if self._doe2 is None:
        self._doe2 = Room2DDoe2Properties(self.host)
    return self._doe2


# add doe2 property methods to the Properties classes
ModelProperties.doe2 = property(model_doe2_properties)
Room2DProperties.doe2 = property(room2d_doe2_properties)

# add writers to the honeybee-core modules
model_writer.inp = model_to_inp
